﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SunFP
{
    class SunFp
    {
        /* 32Bit with 16bit integer and 16bit fractional part */
        private int internalValue;

        private static readonly int NUM_FRAC_BITS = 16;
        private static readonly int NUM_INT_BITS = 32 - NUM_FRAC_BITS;

        private static readonly int MASK_FRAC_BITS = 0x0000FFFF;
        private static readonly UInt32 MASK_INT_BITS = 0xFFFF0000;

        private static readonly int SUNFP_ONE = (1 << 16);
        private static readonly int SUNFP_HALF = (1 << 15);

        private static readonly SunFp SUNFP_PI = new SunFp("3.15");
        private static readonly SunFp SUNFP_PIHALF = new SunFp("1.57");

        private static readonly SunFp[] LogNDividerLookTable = new SunFp[]
        { 
            new SunFp(0),           // index: 0 -> LogN(1)
            new SunFp(0.6931f),     // index: 1 -> LogN(2)
            new SunFp(1.3863f),     // index: 2 -> LogN(4)
            new SunFp(2.0794f),     // index: 3 -> LogN(8)
            new SunFp(2.7726f),     // index: 4 -> LogN(16)
            new SunFp(3.4657f),     // index: 5 -> LogN(32)
            new SunFp(4.1589f),     // index: 6 -> LogN(64)
            new SunFp(4.8520f),     // index: 7 -> LogN(128)
            new SunFp(5.5452f),     // index: 8 -> LogN(256)
            new SunFp(6.2383f),     // index: 9 -> LogN(512)
            new SunFp(6.9315f),     // index: 10 -> LogN(1024)
            new SunFp(7.6246f),     // index: 11 -> LogN(2048)
            new SunFp(8.3178f),     // index: 12 -> LogN(4096)
            new SunFp(9.0109f),     // index: 13 -> LogN(8192)
            new SunFp(9.7041f),     // index: 14 -> LogN(16384)
            new SunFp(10.3972f),     // index: 15 -> LogN(32768)
        };


        public SunFp() { }

        public SunFp(int x)
        {
            internalValue = (x << NUM_FRAC_BITS);
        }

        public SunFp(float x)
        {
            /* Detailed:
            // get integer part of passed value
            int intPart = (int)x;
            // get fractional part of passed value
            float fracPart = x - intPart;
            // convert the integer part to fixed point
            internalValue = intPart << 16;
            // convert the fractional part to fixed point and add it to converted integer part
            internalValue += (int)(SUNFP_ONE * fracPart);
            */

            /* the same in a short way */
            internalValue = (int)(x * SUNFP_ONE);
        }

        public static SunFp GetFpFromInternalValue(long internalValue)
        {
            SunFp fp = new SunFp();
            fp.internalValue = (int)internalValue;
            return fp;
        }

        public SunFp(string s)
        {
            s = s.Trim();
            if (s.Contains('.'))
            {
                string[] parts = s.Split(new char[] {'.'}, 2);

                // get integer part
                internalValue = GetIntFromStr(parts[0]) << NUM_FRAC_BITS;

                // get fractional part
                if (parts.Length > 1)
                {
                    int numFracDigits = parts[1].Length;
                    int fracInt = GetIntFromStr(parts[1]);

                    fracInt = fracInt * SUNFP_ONE / (int)Math.Pow(10, numFracDigits);
                    if (internalValue < 0)
                    {
                        internalValue -= SUNFP_ONE;
                        fracInt = SUNFP_ONE - fracInt;
                    }

                    internalValue |= (fracInt & MASK_FRAC_BITS);

                }
            }
            else
            {
                /* no fractional part, try to parse it as int */
                internalValue = GetIntFromStr(s) << NUM_FRAC_BITS;
            }
        }

        private int GetIntFromStr(string s)
        {
            int intValue;
            if (int.TryParse(s, out intValue))
            {
                return intValue;
            }
            else
            {
                throw new ArgumentException(s + "is no valid number.");
            }
        }

        public string ToStringUsingFloat()
        {
            return (internalValue / (float)SUNFP_ONE).ToString();
        }

        public static string ToString(SunFp value)
        {
            return value.ToString();
        }


        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            int fpFrac = (internalValue & MASK_FRAC_BITS);
            int fpInt = (int)(internalValue & MASK_INT_BITS);

            bool isNegativeWithFracPart = false;
            if (fpInt < 0 && fpFrac != 0)
            {
                fpInt += SUNFP_ONE;
                isNegativeWithFracPart = true;
                if (fpInt == 0)
                {
                    sb.Append('-');
                }
            }

            sb.Append(fpInt >> NUM_FRAC_BITS).ToString();

            // check if fractional part exists
            if (fpFrac != 0)
            {
                sb.Append('.');

                if (isNegativeWithFracPart)
                {
                    fpFrac = SUNFP_ONE - fpFrac;
                }

                while (fpFrac > 0)
                {
                    fpFrac *= 10;
                    sb.Append((char)('0' + (fpFrac >> NUM_FRAC_BITS)));
                    fpFrac &= ((1 << NUM_FRAC_BITS) - 1);
                }

            }

            return sb.ToString();

        }

        public string ToInternalBinString()
        {
            StringBuilder sb = new StringBuilder();

            for (int i = 31; i >= 0; i--)
            {
                if ((internalValue & (1 << i)) != 0)
                {
                    sb.Append('1');
                }
                else
                {
                    sb.Append('0');
                }

                if (i == 16) sb.Append('.');
                else if (i % 4 == 0) sb.Append(' ');
            }

            return sb.ToString();
        }

        public string ToInternalHexString()
        {
            return String.Format("0x{0:X4}", internalValue);
        }

        public float ToFloat()
        {
            return internalValue / (float)SUNFP_ONE;
        }

        public int ToInt()
        {
            return internalValue >> 16;
        }



        /// <summary>
        /// Conversion int -> SunFp
        /// </summary>
        /// <param name="x"></param>
        /// <returns></returns>
        public static implicit operator SunFp(int x)
        {
            SunFp fp = new SunFp(x);
            return fp;
        }

        /// <summary>
        /// Conversion float -> SunFp
        /// </summary>
        /// <param name="x"></param>
        /// <returns></returns>
        public static implicit operator SunFp(float x)
        {
            SunFp fp = new SunFp(x);
            return fp;
        }

        public static SunFp Add(SunFp value1, SunFp value2)
        {
            int v1 = value1.internalValue;
            int v2 = value2.internalValue;

            SunFp v3 = 0;
            v3.internalValue = v1 + v2;
            return v3;
        }

        public static SunFp Sub(SunFp value1, SunFp value2)
        {
            int v1 = value1.internalValue;
            int v2 = value2.internalValue;

            SunFp v3 = 0;
            v3.internalValue = v1 - v2;
            return v3;
        }

        public static SunFp Mul(SunFp value1, SunFp value2)
        {
            int intPart1 = (value1.internalValue) >> 16;
            int intPart2 = (value2.internalValue) >> 16;

            int fracPart1 = (value1.internalValue & MASK_FRAC_BITS);
            int fracPart2 = (value2.internalValue & MASK_FRAC_BITS);

            int res = 0;

            res += (intPart1 * intPart2) << 16;
            res += (intPart1 * fracPart2);
            res += (fracPart1 * intPart2);
            res += ((fracPart1 * fracPart2) >> 16) & MASK_FRAC_BITS;

            SunFp v3 = 0;
            v3.internalValue = res;
            return v3;
        }

        public static SunFp Mul2(SunFp value1, SunFp value2)
        {
            long v1 = value1.internalValue;
            long v2 = value2.internalValue;

            SunFp v3 = 0;
            v3.internalValue = (int)((long)(v1 * v2) >> 16);
            return v3;
        }

        public static SunFp Floor(SunFp value)
        {
            SunFp fpRet = 0;
            fpRet.internalValue = (int)(value.internalValue & MASK_INT_BITS);
            return fpRet;
        }

        public static SunFp Ceil(SunFp value)
        {
            SunFp fpRet = 0;
            if ((value.internalValue & MASK_FRAC_BITS) == 0)
            {
                fpRet.internalValue = (int)(value.internalValue & MASK_INT_BITS);
            }
            else
            {
                fpRet.internalValue = (int)((value.internalValue + SUNFP_ONE) & MASK_INT_BITS);
            }
            return fpRet;
        }

        public static SunFp Round(SunFp value)
        {
            SunFp fpRet = 0;

            fpRet.internalValue += value.internalValue + SUNFP_HALF;

            fpRet = SunFp.Floor(fpRet);
            return fpRet;
        }

        public static SunFp Abs(SunFp value)
        {
            if (value >= 0)
                return value;
            else
                return 0 - value;
        }

        public enum ERoundMidpoint { SMALLER, LARGER };
        public static SunFp RoundMidpoint(SunFp value, ERoundMidpoint type)
        {
            SunFp fpRet = 0;

            if ((value.internalValue & MASK_FRAC_BITS) == SUNFP_HALF)
            {
                fpRet.internalValue += value.internalValue + SUNFP_HALF;
                if (type == ERoundMidpoint.SMALLER)
                {
                    fpRet.internalValue -= SUNFP_ONE;
                }
                
            }
            else
            {
                fpRet.internalValue += value.internalValue + SUNFP_HALF;
            }

            fpRet = SunFp.Floor(fpRet);
            return fpRet;
        }

        public static SunFp Div2(SunFp value1, SunFp value2)
        {
            int v1 = value1.internalValue;
            int v2 = value2.internalValue;

            /* calculate the reciprocal 1 / value2.
             * This is done by (1 << 32) / value2.
             * Because 1 << 32 does not fit into a 32bit value, we would need a long here.
             * To avoid this (with the cost of additional error) we use 1 << 31 and after multiplication,
             * the result is shifted 1 by left. */
            uint v2reciprocal = 1;
            v2reciprocal <<= 31;
            v2reciprocal = (uint)(v2reciprocal / v2);

            SunFp v2ReciprocalFp = SunFp.GetFpFromInternalValue(v2reciprocal);

            SunFp v3 = SunFp.Mul(value1, v2ReciprocalFp);
            v3.internalValue <<= 1;
            return v3;
        }

        public static SunFp Div(SunFp value1, SunFp value2)
        {
            long v1 = value1.internalValue;
            long v2 = value2.internalValue;

            SunFp v3 = 0;
            v3.internalValue = (int)((long)(v1 << 16) / v2);
            return v3;
        }

        public static SunFp Sqrt(SunFp v)
        {
            SunFp fpRet = 1;
            SunFp error = 0.001f;

            int iteration = 0;
            while ((iteration++ < 10) && (SunFp.Abs((fpRet * fpRet) - v) >= error) )
            {
                fpRet = 0.5f * (fpRet + (v / fpRet));
            }

            return fpRet;
        }

        public static SunFp Power(SunFp v, int n)
        {
            SunFp fpRet = 1;

            for (int i = 1; i <= n; i++)
            {
                fpRet *= v;
            }
            return fpRet;
        }

        /// <summary>
        /// Calculates x^y:
        /// x^y = e^(ln(x^y)) = e^(y * ln(x))
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public static SunFp Power(SunFp x, SunFp y)
        {
            return Exp(y * LogN_Taylor(x));
        }

        public static SunFp Factorial(int n)
        {
            int val = 1;

            for (int i = 1; i <= n; i++)
            {
                val *= i;
            }

            return val;
        }

        /// <summary>
        /// Calculates approximate sinus using taylor series:
        /// sin(x) = x - x^3/3! + x^5/5! - x^7/7! ...
        /// </summary>
        /// <param name="v">Radian value to calcuate sinus from, must be in range [-PI, PI].</param>
        /// <returns></returns>
        public static SunFp Sin(SunFp v)
        {
            SunFp fpRet = v;
            SunFp tmp;
            SunFp fac7 = Factorial(7);

            if (v < -SUNFP_PI || v > SUNFP_PI)
            {
                throw new ArgumentException("Sin: Value out of range [-PI, PI]");
            }

            tmp = Power(v, 3) / Factorial(3);
            fpRet = fpRet - tmp;

            tmp = Power(v, 5) / Factorial(5);
            fpRet = fpRet + tmp;

            tmp = Power(v, 7) / fac7;
            fpRet = fpRet - tmp;

            /* 9! will overflow the 16bit integer part, so calc step by step */
            tmp = Power(v, 9) / fac7;
            tmp = tmp / (9 * 8);
            fpRet = fpRet + tmp;

            return fpRet;
        }

        /// <summary>
        /// Calculates approximate cosinus using taylor series:
        /// cos(x) = x - x^2/2! + x^4/4! - x^6/6! + x^8/8!.
        /// </summary>
        /// <param name="v">Radian value to calcuate cosinus from, must be in range [-PI, PI].</param>
        /// <returns></returns>
        public static SunFp Cos(SunFp v)
        {
            SunFp fpRet = 1;
            SunFp tmp;

            if (v < -SUNFP_PI || v > SUNFP_PI)
            {
                throw new ArgumentException("Sin: Value out of range [-PI, PI]");
            }

            tmp = Power(v, 2) / Factorial(2);
            fpRet = fpRet - tmp;

            tmp = Power(v, 4) / Factorial(4);
            fpRet = fpRet + tmp;

            tmp = Power(v, 6) / Factorial(6);
            fpRet = fpRet - tmp;

            tmp = Power(v, 8) / Factorial(7); /* avoid overflow */
            tmp = tmp / 8;
            fpRet = fpRet + tmp;

            tmp = Power(v, 10) / Factorial(7);
            tmp = tmp / (8*9*10);
            fpRet = fpRet - tmp;

            return fpRet;
        }

        /// <summary>
        /// Calculates approximate tan by sin(x) / cos(x).
        /// </summary>
        /// <param name="v">Radian value to calcuate tangens from, must be in range (-PI/2, PI/2).</param>
        /// <returns></returns>
        public static SunFp Tan(SunFp v)
        {
            if (v <= -SUNFP_PIHALF || v >= SUNFP_PIHALF)
            {
                throw new ArgumentException("Sin: Value out of range (-PI/2, PI/2)");
            }

            return Sin(v) / Cos(v);
        }

        /// <summary>
        /// Calculates approximate sinus hyperbolicus by taylor series:
        /// sinh(x) = x + x^3/3! + x^5/5! + x^7/7!.
        /// </summary>
        /// <param name="v">Radian value to calcuate sinus hyperbolicus from, must be in range [-PI, PI].</param>
        /// <returns></returns>
        public static SunFp Sinh(SunFp v)
        {
            SunFp fpRet =v;
            SunFp tmp;

            if (v < -SUNFP_PI || v > SUNFP_PI)
            {
                throw new ArgumentException("Sinh: Value out of range [-PI, PI]");
            }

            tmp = Power(v, 3) / Factorial(3);
            fpRet = fpRet + tmp;
            tmp = Power(v, 5) / Factorial(5);
            fpRet = fpRet + tmp;
            tmp = Power(v, 7) / Factorial(7);
            fpRet = fpRet + tmp;

            return fpRet;
        }

        /// <summary>
        /// Calculates approximate cosinus hyperbolicus by taylor series:
        /// sinh(x) = x + x^2/2! + x^4/4! + x^6/6!.
        /// </summary>
        /// <param name="v">Radian value to calcuate cosinus hyperbolicus from, must be in range [-PI, PI].</param>
        /// <returns></returns>
        public static SunFp Cosh(SunFp v)
        {
            SunFp fpRet = 1;
            SunFp tmp;

            if (v < -SUNFP_PI || v > SUNFP_PI)
            {
                throw new ArgumentException("Cosh: Value out of range [-PI, PI]");
            }

            tmp = Power(v, 2) / Factorial(2);
            fpRet = fpRet + tmp;
            tmp = Power(v, 4) / Factorial(4);
            fpRet = fpRet + tmp;
            tmp = Power(v, 6) / Factorial(6);
            fpRet = fpRet + tmp;

            tmp = Power(v, 8) / Factorial(6);
            tmp = tmp / 7;
            tmp = tmp / 8;
            fpRet = fpRet + tmp;

            return fpRet;
        }

        public static SunFp Tanh(SunFp v)
        {
            return Sinh(v) / Cosh(v);
        }

        /// <summary>
        /// Calculates the exponential of e to the given power v.
        /// e^x = 1 + x + x^2/2! + x^3/3! + ...
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static SunFp Exp(SunFp v)
        {
            SunFp fpRet;
            SunFp tmp = 1;

            fpRet = 1;

            for (int iteration = 1; iteration <= 14; iteration++)
            {
                tmp = 1;
                for (int i = 1; i <= iteration; i++)
                {
                    tmp = tmp * (v / i);
                }
                fpRet = fpRet + tmp;
            }

            return fpRet;
        }

        /// <summary>
        /// Inverse hyperbolic tangent
        /// Input range: (-1, 1)
        /// Arctanh = x + x^3/3 + x^5/5 + x^7/7 + ...
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static SunFp ArcTanh(SunFp v)
        {
            SunFp fpRet = v;
            SunFp res = v;

            if (v <= -1 || v >= 1)
            {
                throw new ArgumentException("ArcTanh: Value out of range (-1, 1)");
            }

            res += (Power(v, 3) / 3);
            res += (Power(v, 5) / 5);
            res += (Power(v, 7) / 7);
            res += (Power(v, 9) / 9);
            res += (Power(v, 11) / 11);

            return res;
        }

        /// <summary>
        /// log(x) = 2 * artanh( (x-1) / (x+1))
        /// Input range: [0, +inf]
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static SunFp LogN_ArcTanH(SunFp v)
        {
            if (v < 0)
            {
                throw new ArgumentException("LogN_ArcTanH: Value out of range [0, +inf]");
            }

            SunFp res = 2;

            res *= (SunFp.ArcTanh((v - 1) / (v + 1)));
            return res;
        }

        /// <summary>
        /// Taylor series:
        /// log(1+x) = x - x^2/2 + x^3/3 - x^4/4 + x^5/5 + ..., only valid for input range (-1, 1)
        /// Input range: (0, 2^16-1)
        /// The input value is subtracted by 1 to get input range (0, 2).
        /// Scaling is used to handle all input values.
        /// If e.g v is 25, then divider is calculated to be 16 to be in input range (0,2):
        /// log(25) = log(25/16 * 16) = log(25/16) + log(16).
        /// The log values for input 2^x are precalcluated in a lookup table.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static SunFp LogN_Taylor(SunFp v)
        {
            int dividerLogLoopUpTableIdx;

            if (v < 0)
            {
                throw new ArgumentException("LogN_Taylor: Value out of range [0, +inf]");
            }

            SunFp divider = 1;
            dividerLogLoopUpTableIdx = 0;

            /* find divider so that v >= divider and v < divider + 1 */
            SunFp lowerCmp = 2;
            for (int exp = 1; exp < LogNDividerLookTable.Length; exp++)
            {
                SunFp upperCmp = lowerCmp * 2;

                if (upperCmp == new SunFp(-32768))
                {
                    /* 2^16 is 32768 due to overflow, subtract one to get 32767 */
                    upperCmp -= 1;
                }
                if (v >= lowerCmp && v <= upperCmp)
                {
                    divider = lowerCmp;
                    dividerLogLoopUpTableIdx = exp;
                    break;
                }
                lowerCmp = upperCmp;
            }

            v = v / divider;
            v = v - 1;

            // taylor series
            SunFp res = v;
            res -= (Power(v, 2) / 2);
            res += (Power(v, 3) / 3);
            res -= (Power(v, 4) / 4);
            res += (Power(v, 5) / 5);
            res -= (Power(v, 6) / 6);
            res += (Power(v, 7) / 7);

            // add precalculated 2^x logarithm
            res += LogNDividerLookTable[dividerLogLoopUpTableIdx];

            return res;
        }

        public static SunFp operator +(SunFp v1, SunFp v2)
        {
            return SunFp.Add(v1, v2);
        }

        public static SunFp operator - (SunFp v1, SunFp v2)
        {
            return SunFp.Sub(v1, v2);
        }

        public static SunFp operator * (SunFp v1, SunFp v2)
        {
            return SunFp.Mul(v1, v2);
        }

        public static SunFp operator / (SunFp v1, SunFp v2)
        {
            return SunFp.Div(v1, v2);
        }

        public static bool operator < (SunFp v1, SunFp v2)
        {
            return v1.internalValue < v2.internalValue;
        }

        public static bool operator <= (SunFp v1, SunFp v2)
        {
            return v1.internalValue <= v2.internalValue;
        }

        public static bool operator > (SunFp v1, SunFp v2)
        {
            return v1.internalValue > v2.internalValue;
        }

        public static bool operator >= (SunFp v1, SunFp v2)
        {
            return v1.internalValue >= v2.internalValue;
        }

        public static bool operator ==(SunFp v1, SunFp v2)
        {
            return v1.internalValue == v2.internalValue;
        }

        public static bool operator !=(SunFp v1, SunFp v2)
        {
            return v1.internalValue != v2.internalValue;
        }

        public static SunFp operator -(SunFp v1)
        {
            return 0 - v1;
        }

        public override bool Equals(System.Object obj)
        {
            if (obj == null)
            {
                return false;
            }

            SunFp p = obj as SunFp;
            if ((System.Object)p == null)
            {
                return false;
            }

            // Return true if the fields match:
            return (this.internalValue == p.internalValue);
        }

        public override int GetHashCode()
        {
            return internalValue;
        }

    }
}
